"""
Predefined kernels used for convolution by arcgis.raster.functions.convolution()
in addition to user defined kernels
"""

LINE_DETECTION_HORIZONTAL = 0
LINE_DETECTION_VERTICAL = 1
LINE_DETECTION_LEFT_DIAGONAL = 2
LINE_DETECTION_RIGHT_DIAGONAL = 3
GRADIENT_NORTH = 4
GRADIENT_WEST = 5
GRADIENT_EAST = 6
GRADIENT_SOUTH = 7
GRADIENT_NORTH_EAST = 8
GRADIENT_NORTH_WEST = 9
SMOOTH_ARITHMETIC_MEAN = 10
SMOOTHING_3X3 = 11
SMOOTHING_5X5 = 12
SHARPENING_3X3 = 13
SHARPENING_5X5 = 14
LAPLACIAN_3X3 = 15
LAPLACIAN_5X5 = 16
SOBEL_HORIZONTAL = 17
SOBEL_VERTICAL = 18
SHARPEN = 19
SHARPEN2 = 20
POINT_SPREAD = 21
